#include <map>
#include <vector>
#include <string>
#include <dlfcn.h>

#include "dynpool.hh"
#include "autoptr"

#include <cstdio>

dynlib::dynlib(const char *fn) : handle(dlopen(fn, RTLD_LAZY)), libname(fn)
{
	std::printf("Loading lib '%s'\n", libname);
}

dynlib::~dynlib()
{
	/* Automatically close the lib when nothing refers to it */
	if(handle)dlclose(handle);
	std::printf("Unloading lib '%s' (it is no longer in use)\n", libname);
}

void *dynlib::getsym(const char *sym)
{
	if(!handle)return NULL;
	return dlsym(handle, sym);
}

void dynclass::operator=(const dyncontent &p)
{
	dyncontent::operator=(p);
	std::printf("Loading sym '%s'\n", name);
}
	
dynclass::dynclass(autoptr<dynlib> &l) : lib(l)
{
}

dynclass::~dynclass()
{
	std::printf("Unloading sym '%s'\n", name);
}

class dyn_NameList : public std::map<std::string, autoptr<dynclass> >
{
};

dynpool::dynpool() : hash(new dyn_NameList)
{
}

dynpool::~dynpool()
{
	delete hash; hash = NULL;
}

void dynpool::Exception(const dynclass &old, const dyncontent &prob)
{
	std::fprintf(stderr, "Paha paha ongelma.\n");
	throw;
}

void dynpool::Exception(const char *dllname)
{
	std::fprintf(stderr, "Paha paha ongelma.\n");
	throw;
}

void dynpool::Exception(const char *dllname, const char *symbol)
{
	std::fprintf(stderr, "Paha paha ongelma.\n");
	throw;
}

void dynpool::load(const char *fn)
{
	autoptr<dynlib> lib = new dynlib(fn);
	
	if(!lib->ok())
	{
		Exception(fn);
	}
	
	const char *dlerr;
	struct dyncontent *contents;
	contents = (struct dyncontent *)lib->getsym("contents");
	if((dlerr = dlerror()))
	{
		Exception(fn, dlerr);
		return;
	}
	
	for(unsigned a=0; contents[a].name; ++a)
	{
		dyn_NameList::const_iterator i;
		i = hash->find(contents[a].name);
		
		if(i != hash->end())
		{
    		/* Symbol already loaded! */
    		
    		Exception(*(i->second), contents[a]);
    		return;
		}
		
		dynclass *rec = new dynclass(lib);
		*rec = contents[a];
		(*hash)[contents[a].name] = rec;
	}
}

unsigned dynpool::size() const
{
	return hash->size();
}

const dynpool::ent *dynpool::find(const char *name) const
{
	dyn_NameList::const_iterator i = hash->find(name);
	if(i != hash->end())return i->second;
	return NULL;
}

void dynpool::unregister(const char *name)
{
	dyn_NameList::iterator i = hash->find(name);
	if(i != hash->end())
		hash->erase(i);
	else
		std::printf("Tried to unregister symbol '%s', which was not in the pool.\n", name);
}

const dynpool::ent &dynpool::operator[] (unsigned ind) const
{
	/* OUCH! uff. This is not beautiful! */
	dyn_NameList::const_iterator i = hash->begin();
	for(unsigned a=0; a<ind && i!=hash->end(); ++a)
		++i;
	
	return *i->second;
}
